LIBRARY		"libwinpr-interlocked"
EXPORTS
	InterlockedCompareExchange64		@1
	InitializeListHead			@2
	IsListEmpty				@3
	RemoveEntryList				@4
	InsertHeadList				@5
	RemoveHeadList				@6
	InsertTailList				@7
	RemoveTailList				@8
	AppendTailList				@9
	PushEntryList				@10
	PopEntryList				@11
