/**
 * WinPR: Windows Portable Runtime
 * Credential Security Support Provider (CredSSP)
 *
 * Copyright 2010-2012 Marc-Andre Moreau <marcandre.moreau@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WINPR_SSPI_CREDSSP_PRIVATE_H
#define WINPR_SSPI_CREDSSP_PRIVATE_H

#include <winpr/sspi.h>

#include "../sspi.h"

struct _CREDSSP_CONTEXT
{
	BOOL server;
};
typedef struct _CREDSSP_CONTEXT CREDSSP_CONTEXT;

CREDSSP_CONTEXT* credssp_ContextNew(void);
void credssp_ContextFree(CREDSSP_CONTEXT* context);

#endif /* WINPR_SSPI_CREDSSP_PRIVATE_H */
